@@ -18,7 +18,7 @@
 */
 
 /*
- * $Id: ccid_usb.c 2359 2007-01-27 13:31:33Z rousseau $
+ * $Id: ccid_usb.c 2510 2007-05-03 19:21:54Z rousseau $
  */
 
 #define __CCID_USB__
@@ -82,7 +82,7 @@ typedef struct
 /* The _usbDevice structure must be defined before including ccid_usb.h */
 #include "ccid_usb.h"
 
-static int get_end_points(struct usb_device *dev, _usbDevice *usb_device);
+static int get_end_points(struct usb_device *dev, _usbDevice *usbdevice);
 int ccid_check_firmware(struct usb_device *dev);
 static unsigned int *get_data_rates(unsigned int reader_index,
 	struct usb_device *dev);
@@ -428,6 +428,7 @@ status_t OpenUSBByName(unsigned int reader_index, /*@null@*/ char *device)
 					usbDevice[reader_index].ccid.bCurrentSlotIndex = 0;
 					usbDevice[reader_index].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT;
 					usbDevice[reader_index].ccid.arrayOfSupportedDataRates = get_data_rates(reader_index, dev);
+					usbDevice[reader_index].ccid.bInterfaceProtocol = usb_interface->altsetting->bInterfaceProtocol;
 					goto end;
 				}
 			}
@@ -596,7 +597,7 @@ _ccid_descriptor *get_ccid_descriptor(unsigned int reader_index)
  *					get_end_points
  *
  ****************************************************************************/
-static int get_end_points(struct usb_device *dev, _usbDevice *usb_device)
+static int get_end_points(struct usb_device *dev, _usbDevice *usbdevice)
 {
 	int i;
 	int bEndpointAddress;
@@ -605,7 +606,7 @@ static int get_end_points(struct usb_device *dev, _usbDevice *usb_device)
 	/*
 	 * 3 Endpoints maximum: Interrupt In, Bulk In, Bulk Out
 	 */
-	for (i=0; i<3; i++)
+	for (i=0; i<usb_interface->altsetting->bNumEndpoints; i++)
 	{
 		if (usb_interface->altsetting->endpoint[i].bmAttributes != USB_ENDPOINT_TYPE_BULK)
 			continue;
@@ -613,10 +614,10 @@ static int get_end_points(struct usb_device *dev, _usbDevice *usb_device)
 		bEndpointAddress = usb_interface->altsetting->endpoint[i].bEndpointAddress;
 
 		if ((bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_IN)
-			usb_device->bulk_in = bEndpointAddress;
+			usbdevice->bulk_in = bEndpointAddress;
 
 		if ((bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_ENDPOINT_OUT)
-			usb_device->bulk_out = bEndpointAddress;
+			usbdevice->bulk_out = bEndpointAddress;
 	}
 
 	return 0;
@@ -663,6 +664,9 @@ static int get_end_points(struct usb_device *dev, _usbDevice *usb_device)
 				usb_interface->altsetting->extralen = 54;
 				usb_interface->altsetting->extra =
 					usb_interface->altsetting->endpoint[i].extra;
+				/* avoid double free on close */
+				usb_interface->altsetting->endpoint[i].extra = NULL;
+				usb_interface->altsetting->endpoint[i].extralen = 0;
 				break;
 			}
 		}
@@ -727,14 +731,11 @@ static unsigned int *get_data_rates(unsigned int reader_index,
 	unsigned int *int_array;
 
 	/* See CCID 3.7.3 page 25 */
-	n = usb_control_msg(usbDevice[reader_index].handle,
+	n = ControlUSB(reader_index,
 		0xA1, /* request type */
 		0x03, /* GET_DATA_RATES */
 		0x00, /* value */
-		usbDevice[reader_index].interface, /* interface */
-		(char *)buffer,
-		sizeof(buffer),
-		usbDevice[reader_index].ccid.readTimeout * 1000);
+		buffer, sizeof(buffer));
 
 	/* we got an error? */
 	if (n <= 0)
@@ -785,3 +786,23 @@ static unsigned int *get_data_rates(unsigned int reader_index,
 	return int_array;
 }
 
+int ControlUSB(int reader_index, int requesttype, int request, int value,
+	unsigned char *bytes, unsigned int size)
+{
+	int ret;
+
+	DEBUG_COMM2("request: 0x%02X", request);
+
+	if (0 == (requesttype & 0x80))
+		DEBUG_XXD("send: ", bytes, size);
+
+	ret = usb_control_msg(usbDevice[reader_index].handle, requesttype,
+		request, value, usbDevice[reader_index].interface, (char *)bytes, size,
+		usbDevice[reader_index].ccid.readTimeout * 1000);
+
+	if (requesttype & 0x80)
+		DEBUG_XXD("receive: ", bytes, ret);
+
+	return ret;
+}
+
