@@ -18,11 +18,12 @@
 */
 
 /*
- * $Id: commands.c 2358 2007-01-27 13:28:26Z rousseau $
+ * $Id: commands.c 2511 2007-05-07 09:31:13Z rousseau $
  */
 
 #include <string.h>
 #include <stdlib.h>
+#include <errno.h>
 #include <pcsclite.h>
 #include <ifdhandler.h>
 #include <reader.h>
@@ -34,11 +35,17 @@
 #include "ccid_ifdhandler.h"
 #include "config.h"
 #include "debug.h"
+#include "ccid_usb.h"
 
 /* All the pinpad readers I used are more or less bogus
  * I use code to change the user command and make the firmware happy */
 #define BOGUS_PINPAD_FIRMWARE
 
+/* The firmware of SCM readers reports dwMaxCCIDMessageLength = 263
+ * instead of 270 so this prevents from sending a full length APDU
+ * of 260 bytes since the driver check this value */
+#define BOGUS_SCM_FIRMWARE_FOR_dwMaxCCIDMessageLength
+
 #define max( a, b )   ( ( ( a ) > ( b ) ) ? ( a ) : ( b ) )
 #define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
 #define IFD_ERROR_INSUFFICIENT_BUFFER 700
@@ -78,6 +85,84 @@ RESPONSECODE CmdPowerOn(unsigned int reader_index, unsigned int * nlength,
 	RESPONSECODE return_value = IFD_SUCCESS;
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 
+	if (ICCD_A == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+		unsigned char pcbuffer[SIZE_GET_SLOT_STATUS];
+
+		/* first power off to reset the ICC state machine */
+		r = CmdPowerOff(reader_index);
+		if (r != IFD_SUCCESS)
+			return r;
+
+		/* wait for ready */
+		r = CmdGetSlotStatus(reader_index, pcbuffer);
+		if (r != IFD_SUCCESS)
+			return r;
+
+		/* Power On */
+		r = ControlUSB(reader_index, 0xA1, 0x62, 0, buffer, *nlength);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Power On failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		*nlength = r;
+
+		return IFD_SUCCESS;
+	}
+
+	if (ICCD_B == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+		unsigned char tmp[MAX_ATR_SIZE+1];
+
+		/* first power off to reset the ICC state machine */
+		r = CmdPowerOff(reader_index);
+		if (r != IFD_SUCCESS)
+			return r;
+
+		/* Power On */
+		r = ControlUSB(reader_index, 0x21, 0x62, 1, NULL, 0);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Power On failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		/* Data Block */
+		r = ControlUSB(reader_index, 0xA1, 0x6F, 0, tmp, sizeof(tmp));
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Data Block failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		if (tmp[0] != 0x00)
+		{
+			DEBUG_CRITICAL2("bResponseType: 0x%02X", tmp[0]);
+
+			/* Status Information? */
+			if (0x40 == tmp[0])
+				ccid_error(tmp[2], __FILE__, __LINE__, __FUNCTION__);
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		DEBUG_INFO_XXD("Data Block: ", tmp, r);
+		if (*nlength > r-1)
+			*nlength = r-1;
+		memcpy(buffer, tmp+1, *nlength);
+
+		return IFD_SUCCESS;
+	}
+
 	/* store length of buffer[] */
 	length = *nlength;
 
@@ -118,12 +203,12 @@ again:
 			((GEMPC433 == ccid_descriptor->readerID)
 			|| (CHERRYXX33 == ccid_descriptor->readerID)))
 		{
-			unsigned char cmd[] = {0x1F, 0x01};
-			unsigned char res[1];
-			unsigned int res_length = sizeof(res);
+			unsigned char cmd_tmp[] = {0x1F, 0x01};
+			unsigned char res_tmp[1];
+			unsigned int res_length = sizeof(res_tmp);
 
-			if ((return_value = CmdEscape(reader_index, cmd, sizeof(cmd), res,
-				&res_length)) != IFD_SUCCESS)
+			if ((return_value = CmdEscape(reader_index, cmd_tmp,
+				sizeof(cmd_tmp), res_tmp, &res_length)) != IFD_SUCCESS)
 				return return_value;
 
 			/* avoid looping if we can't switch mode */
@@ -188,21 +273,18 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 	if (TxLength > 19+CMD_BUF_SIZE) /* command too large? */
 	{
 		DEBUG_INFO3("Command too long: %d > %d", TxLength, 19+CMD_BUF_SIZE);
-		*RxLength = 0;
 		return IFD_NOT_SUPPORTED;
 	}
 
 	if (TxLength < 19+4 /* 4 = APDU size */)	/* command too short? */
 	{
 		DEBUG_INFO3("Command too short: %d < %d", TxLength, 19+4);
-		*RxLength = 0;
 		return IFD_NOT_SUPPORTED;
 	}
 
 	if (dw2i(TxBuffer, 15) + 19 != TxLength) /* ulDataLength field coherency */
 	{
 		DEBUG_INFO3("Wrong lengths: %d %d", dw2i(TxBuffer, 15) + 19, TxLength);
-		*RxLength = 0;
 		return IFD_NOT_SUPPORTED;
 	}
 
@@ -288,16 +370,16 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 	if ((SPR532 == ccid_descriptor->readerID) && (TxBuffer[15] == 4))
 	{
 		RESPONSECODE return_value;
-		unsigned char cmd[] = { 0x80, 0x02, 0x00 };
-		unsigned char res[1];
-		unsigned int res_length = sizeof(res);
+		unsigned char cmd_tmp[] = { 0x80, 0x02, 0x00 };
+		unsigned char res_tmp[1];
+		unsigned int res_length = sizeof(res_tmp);
 
 		/* the SPR532 will append the PIN code without any padding */
-		return_value = CmdEscape(reader_index, cmd, sizeof(cmd), res,
-			&res_length);
+		return_value = CmdEscape(reader_index, cmd_tmp, sizeof(cmd_tmp),
+			res_tmp, &res_length);
 		if (return_value != IFD_SUCCESS)
 		{
-			ccid_error(res[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);
+			ccid_error(res_tmp[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);
 			return return_value;
 		}
 	}
@@ -308,10 +390,7 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 	ccid_descriptor -> readTimeout = max(30, TxBuffer[0]+10);	/* at least 30 seconds */
 
 	if (WritePort(reader_index, a, cmd) != STATUS_SUCCESS)
-	{
-		*RxLength = 0;
 		return IFD_COMMUNICATION_ERROR;
-	}
 
 	ret = CCID_Receive(reader_index, RxLength, RxBuffer, NULL);
 
@@ -355,6 +434,9 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 	int old_read_timeout;
 	RESPONSECODE ret;
+#ifdef BOGUS_PINPAD_FIRMWARE
+	int bNumberMessages = 0; /* for GemPC Pinpad */
+#endif
 
 	cmd[0] = 0x69;	/* Secure */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
@@ -369,21 +451,18 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	if (TxLength > 24+CMD_BUF_SIZE) /* command too large? */
 	{
 		DEBUG_INFO3("Command too long: %d > %d", TxLength, 24+CMD_BUF_SIZE);
-		*RxLength = 0;
 		return IFD_NOT_SUPPORTED;
 	}
 
 	if (TxLength < 24+4 /* 4 = APDU size */) /* command too short? */
 	{
 		DEBUG_INFO3("Command too short: %d < %d", TxLength, 24+4);
-		*RxLength = 0;
 		return IFD_NOT_SUPPORTED;
 	}
 
 	if (dw2i(TxBuffer, 20) + 24 != TxLength) /* ulDataLength field coherency */
 	{
 		DEBUG_INFO3("Wrong lengths: %d %d", dw2i(TxBuffer, 20) + 24, TxLength);
-		*RxLength = 0;
 		return IFD_NOT_SUPPORTED;
 	}
 
@@ -391,7 +470,6 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	if (TxBuffer[11] > 3)
 	{
 		DEBUG_INFO2("Wrong bNumberMessage: %d", TxBuffer[11]);
-		*RxLength = 0;
 		return IFD_NOT_SUPPORTED;
 	}
 
@@ -441,6 +519,7 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 
 		/* the reader does not support any other value than 3 for the number
 		 * of messages */
+		bNumberMessages = TxBuffer[11];
 		if (0x03 != TxBuffer[11])
 		{
 			DEBUG_INFO2("Correct bNumberMessages for GemPC Pinpad (was %d)",
@@ -519,6 +598,9 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	{
 		cmd[21] = 0x00; /* set bNumberMessages to 0 */
 	}
+
+	if (GEMPCPINPAD == ccid_descriptor->readerID)
+		cmd[21] = bNumberMessages;	/* restore the real value */
 #endif
 
 	/* We know the size of the CCID message now */
@@ -528,10 +610,7 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	ccid_descriptor -> readTimeout = max(30, TxBuffer[0]+10);	/* at least 30 seconds */
 
 	if (WritePort(reader_index, a, cmd) != STATUS_SUCCESS)
-	{
-		*RxLength = 0;
  		return IFD_COMMUNICATION_ERROR;
-	}
 
  	ret = CCID_Receive(reader_index, RxLength, RxBuffer, NULL);
 
@@ -662,6 +741,51 @@ RESPONSECODE CmdPowerOff(unsigned int reader_index)
 	RESPONSECODE return_value = IFD_SUCCESS;
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 
+	if (ICCD_A == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+
+		/* PowerOff */
+		r = ControlUSB(reader_index, 0x21, 0x63, 0, NULL, 0);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Power Off failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		return IFD_SUCCESS;
+	}
+
+	if (ICCD_B == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+		unsigned char buffer[3];
+
+		/* PowerOff */
+		r = ControlUSB(reader_index, 0x21, 0x63, 0, NULL, 0);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Power Off failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		/* SlotStatus */
+		r = ControlUSB(reader_index, 0xA1, 0x81, 0, buffer, sizeof(buffer));
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC SlotStatus failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		return IFD_SUCCESS;
+	}
+
 	cmd[0] = 0x63; /* IccPowerOff */
 	cmd[1] = cmd[2] = cmd[3] = cmd[4] = 0;	/* dwLength */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
@@ -706,6 +830,80 @@ RESPONSECODE CmdGetSlotStatus(unsigned int reader_index, unsigned char buffer[])
 	RESPONSECODE return_value = IFD_SUCCESS;
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 
+	if (ICCD_A == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+		unsigned char status[1];
+
+again_status:
+		/* SlotStatus */
+		r = ControlUSB(reader_index, 0xA1, 0xA0, 0, status, sizeof(status)); 
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Slot Status failed: %s", strerror(errno));
+			if (ENODEV == errno)
+				return IFD_NO_SUCH_DEVICE;
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		/* busy */
+		if (status[0] & 0x40)
+		{
+			DEBUG_INFO2("Busy: 0x%02X", status[0]);
+			usleep(1000 * 10);
+			goto again_status;
+		}
+
+		/* simulate a CCID bStatus */
+		/* present and active by default */
+		buffer[7] = CCID_ICC_PRESENT_ACTIVE;
+
+		/* mute */
+		if (0x80 == status[0])
+			buffer[7] = CCID_ICC_ABSENT;
+
+		/* store the status for CmdXfrBlockCHAR_T0() */
+		buffer[0] = status[0];
+
+		return IFD_SUCCESS;
+	}
+
+	if (ICCD_B == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+		unsigned char buffer_tmp[3];
+
+		/* SlotStatus */
+		r = ControlUSB(reader_index, 0xA1, 0x81, 0, buffer_tmp,
+			sizeof(buffer_tmp));
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Slot Status failed: %s", strerror(errno));
+			if (ENODEV == errno)
+				return IFD_NO_SUCH_DEVICE;
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		/* simulate a CCID bStatus */
+		switch (buffer_tmp[1] & 0x03)
+		{
+			case 0:
+				buffer[7] = CCID_ICC_PRESENT_ACTIVE;
+				break;
+			case 1:
+				buffer[7] = CCID_ICC_PRESENT_INACTIVE;
+				break;
+			case 2:
+			case 3:
+				buffer[7] = CCID_ICC_ABSENT;
+		}
+		return IFD_SUCCESS;
+	}
+
 	cmd[0] = 0x65; /* GetSlotStatus */
 	cmd[1] = cmd[2] = cmd[3] = cmd[4] = 0;	/* dwLength */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
@@ -785,8 +983,6 @@ RESPONSECODE CmdXfrBlock(unsigned int reader_index, unsigned int tx_length,
 		case CCID_CLASS_EXTENDED_APDU:
 			return_value = CmdXfrBlockAPDU_extended(reader_index,
 				tx_length, tx_buffer, rx_length, rx_buffer);
-			if (return_value != IFD_SUCCESS)
-				*rx_length = 0;
 			break;
 
 		case CCID_CLASS_CHARACTER:
@@ -802,7 +998,6 @@ RESPONSECODE CmdXfrBlock(unsigned int reader_index, unsigned int tx_length,
 			break;
 
 		default:
-			*rx_length = 0;
 			return_value = IFD_COMMUNICATION_ERROR;
 	}
 
@@ -822,6 +1017,46 @@ RESPONSECODE CCID_Transmit(unsigned int reader_index, unsigned int tx_length,
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 	status_t ret;
 
+	if (ICCD_A == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+
+		/* Xfr Block */
+		r = ControlUSB(reader_index, 0x21, 0x65, 0, tx_buffer, tx_length);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Xfr Block failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		return IFD_SUCCESS;
+	}
+
+	if (ICCD_B == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+
+		/* nul block so we are chaining */
+		if (NULL == tx_buffer)
+			rx_length = 0x10;	/* bLevelParameter */
+
+		/* Xfr Block */
+		DEBUG_COMM2("chain parameter: %d", rx_length);
+		r = ControlUSB(reader_index, 0x21, 0x65, rx_length << 8, tx_buffer,
+			tx_length);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Xfr Block failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		return IFD_SUCCESS;
+	}
+
 	cmd[0] = 0x6F; /* XfrBlock */
 	i2dw(tx_length, cmd+1);	/* APDU length */
 	cmd[5] = ccid_descriptor->bCurrentSlotIndex;	/* slot number */
@@ -860,14 +1095,103 @@ RESPONSECODE CCID_Receive(unsigned int reader_index, unsigned int *rx_length,
 	unsigned char cmd[10+CMD_BUF_SIZE];	/* CCID + APDU buffer */
 	unsigned int length;
 	RESPONSECODE return_value = IFD_SUCCESS;
+	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 	status_t ret;
 
+	if (ICCD_A == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+
+		/* Data Block */
+		r = ControlUSB(reader_index, 0xA1, 0x6F, 0, rx_buffer, *rx_length);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Data Block failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		return IFD_SUCCESS;
+	}
+
+	if (ICCD_B == ccid_descriptor->bInterfaceProtocol)
+	{
+		int r;
+		unsigned char rx_tmp[4];
+
+		/* read a nul block. buffer need to be at least 4-bytes */
+		if (NULL == rx_buffer)
+		{
+			rx_buffer = rx_tmp;
+			*rx_length = sizeof(rx_tmp);
+		}
+
+time_request_ICCD_B:
+		/* Data Block */
+		r = ControlUSB(reader_index, 0xA1, 0x6F, 0, rx_buffer, *rx_length);
+
+		/* we got an error? */
+		if (r < 0)
+		{
+			DEBUG_INFO2("ICC Data Block failed: %s", strerror(errno));
+			return IFD_COMMUNICATION_ERROR;
+		}
+
+		/* bResponseType */
+		switch (rx_buffer[0])
+		{
+			case 0x00:
+				/* the abData field contains the information created by the
+				 * preceding request */
+				break;
+
+			case 0x40:
+				/* Status Information */
+				ccid_error(rx_buffer[2], __FILE__, __LINE__, __FUNCTION__);
+				return IFD_COMMUNICATION_ERROR;
+
+			case 0x80:
+				/* Polling */
+			{
+				int delay;
+
+				delay = (rx_buffer[2] << 8) + rx_buffer[1];
+				DEBUG_COMM2("Pooling delay: %d", delay);
+
+				if (0 == delay)
+					/* host select the delay */
+					delay = 1;
+				usleep(delay * 1000 * 10);
+				goto time_request_ICCD_B;
+			}
+
+			case 0x01:
+			case 0x02:
+			case 0x03:
+			case 0x10:
+				/* Extended case
+				 * Only valid for Data Block frames */
+				if (chain_parameter)
+					*chain_parameter = rx_buffer[0];
+				break;
+
+			default:
+				DEBUG_CRITICAL2("Unknown bResponseType: 0x%02X", rx_buffer[0]);
+				return IFD_COMMUNICATION_ERROR;
+		}
+
+		memmove(rx_buffer, rx_buffer+1, r-1);
+		*rx_length = r-1;
+
+		return IFD_SUCCESS;
+	}
+
 time_request:
 	length = sizeof(cmd);
 	ret = ReadPort(reader_index, &length, cmd);
 	if (ret != STATUS_SUCCESS)
 	{
-		*rx_length = 0;
 		if (STATUS_NO_SUCH_DEVICE == ret)
 			return IFD_NO_SUCH_DEVICE;
 		return IFD_COMMUNICATION_ERROR;
@@ -876,7 +1200,6 @@ time_request:
 	if (length < STATUS_OFFSET+1)
 	{
 		DEBUG_CRITICAL2("Not enough data received: %d bytes", length);
-		*rx_length = 0;
 		return IFD_COMMUNICATION_ERROR;
 	}
 
@@ -902,11 +1225,9 @@ time_request:
 				return IFD_SUCCESS;
 
 			case 0xFD:	/* Parity error during exchange */
-				*rx_length = 0; /* nothing received */
 				return IFD_PARITY_ERROR;
 
 			default:
-				*rx_length = 0; /* nothing received */
 				return IFD_COMMUNICATION_ERROR;
 		}
 	}
@@ -953,6 +1274,15 @@ static RESPONSECODE CmdXfrBlockAPDU_extended(unsigned int reader_index,
 	unsigned int local_rx_length, received_length;
 	int buffer_overflow = 0;
 
+	if (ICCD_B == ccid_descriptor->bInterfaceProtocol)
+	{
+		/* length is on 16-bits only
+		 * if a size > 0x1000 is used then usb_control_msg() fails with
+		 * "Invalid argument" */
+		if (*rx_length > 0x1000)
+			*rx_length = 0x1000;
+	}
+
 	DEBUG_COMM2("T=0 (extended): %d bytes", tx_length);
 
 	/* send the APDU */
@@ -1087,9 +1417,20 @@ static RESPONSECODE CmdXfrBlockTPDU_T0(unsigned int reader_index,
 	/* command length too big for CCID reader? */
 	if (tx_length > ccid_descriptor->dwMaxCCIDMessageLength-10)
 	{
-		DEBUG_CRITICAL3("Command too long (%d bytes) for max: %d bytes",
+#ifdef BOGUS_SCM_FIRMWARE_FOR_dwMaxCCIDMessageLength
+		if (263 == ccid_descriptor->dwMaxCCIDMessageLength)
+		{
+			DEBUG_INFO3("Command too long (%d bytes) for max: %d bytes."
+				" SCM reader with bogus firmware?",
 				tx_length, ccid_descriptor->dwMaxCCIDMessageLength-10);
-		return IFD_COMMUNICATION_ERROR;
+		}
+		else
+#endif
+		{
+			DEBUG_CRITICAL3("Command too long (%d bytes) for max: %d bytes",
+				tx_length, ccid_descriptor->dwMaxCCIDMessageLength-10);
+			return IFD_COMMUNICATION_ERROR;
+		}
 	}
 
 	/* command length too big for CCID driver? */
@@ -1335,9 +1676,76 @@ static RESPONSECODE CmdXfrBlockCHAR_T0(unsigned int reader_index,
 	unsigned int exp_len, in_len;
 	unsigned char ins, *in_buf;
 	RESPONSECODE return_value = IFD_SUCCESS;
+	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 
 	DEBUG_COMM2("T=0: %d bytes", snd_len);
 
+	if (ICCD_A == ccid_descriptor->bInterfaceProtocol)
+	{
+		unsigned char pcbuffer[SIZE_GET_SLOT_STATUS];
+
+		/* Command to send to the smart card (must be 5 bytes) */
+		memset(cmd, 0, sizeof(cmd));
+		if (snd_len == 4)
+		{
+			memcpy(cmd, snd_buf, 4);
+			snd_buf += 4;
+			snd_len -= 4;
+		}
+		else
+		{
+			memcpy(cmd, snd_buf, 5);
+			snd_buf += 5;
+			snd_len -= 5;
+		}
+
+		/* at most 5 bytes */
+		return_value = CCID_Transmit(reader_index, 5, cmd, 0, 0);
+		if (return_value != IFD_SUCCESS)
+			return return_value;
+
+		/* wait for ready */
+		return_value = CmdGetSlotStatus(reader_index, pcbuffer);
+		if (return_value != IFD_SUCCESS)
+			return return_value;
+
+		if ((0x20 != pcbuffer[0]) && (snd_len > 0))
+		{
+			/* continue sending the APDU */
+			return_value = CCID_Transmit(reader_index, snd_len, snd_buf, 0, 0);
+			if (return_value != IFD_SUCCESS)
+				return return_value;
+		}
+		else
+		{
+			if ((0x20 == pcbuffer[0]) && (*rcv_len > 2))
+				/* we will just read SW1-SW2 */
+				*rcv_len = 2;
+
+			return_value = CCID_Receive(reader_index, rcv_len, rcv_buf, NULL);
+			if (return_value != IFD_SUCCESS)
+				DEBUG_CRITICAL("CCID_Receive failed");
+
+			return return_value;
+		}
+
+		/* wait for ready */
+		return_value = CmdGetSlotStatus(reader_index, pcbuffer);
+		if (return_value != IFD_SUCCESS)
+			return return_value;
+
+		if ((0x20 == pcbuffer[0]) && (*rcv_len > 2))
+			/* we will just read SW1-SW2 */
+			*rcv_len = 2;
+
+		/* read SW1-SW2 */
+		return_value = CCID_Receive(reader_index, rcv_len, rcv_buf, NULL);
+		if (return_value != IFD_SUCCESS)
+			DEBUG_CRITICAL("CCID_Receive failed");
+
+		return return_value;
+	}
+
 	in_buf = tmp_buf;
 	in_len = 0;
 	*rcv_len = 0;
@@ -1473,10 +1881,7 @@ static RESPONSECODE CmdXfrBlockTPDU_T1(unsigned int reader_index,
 		tx_buffer, tx_length, rx_buffer, *rx_length);
 
 	if (ret < 0)
-	{
-		*rx_length = 0;
 		return_value = IFD_COMMUNICATION_ERROR;
-	}
 	else
 		*rx_length = ret;
 
